unit Recording;

interface

uses
  Classes;

type
  TCalcMode = (cmSimple, cmZero, cmEps, cmPartialPivot, cmCompletePivot);
  float = Real; // for example single
  TabFloat = array of float;
  Tab2DFloat = array of array of float;

  TCommandProcessor = class;
  TModel = class;

  TResults = class
  private
    FModel: TModel;
    FSize: integer;
    procedure SetSize(const Value: integer);
  public
    calculated: TabFloat;
    is_calculated: boolean;
    existsExact: boolean;
    exact: TabFloat;
    swaps: array of integer;
    constructor Create(AModel: TModel);
    procedure RestoreOrder;
    procedure CopyExact(Aexact: TabFloat);
    property Size: integer read FSize write SetSize;
  end;

  TModel = class
  private
    FCommandProcessor: TCommandProcessor;
    A_sav: Tab2DFloat;
    b_sav: TabFloat;
    FResults: TResults;
    FSize: integer;
    FChanged: boolean;
    procedure SetSize(const Value: integer);
  public
    A: Tab2DFloat;
    b: TabFloat;
    isSingular: boolean;
    isRankLess: boolean;
    otherException: boolean;
    hasZeroOnDiagonal: boolean;
    constructor Create();
    destructor Destroy(); override;
    procedure Clear();
    procedure Generate();
    procedure Save();
    procedure Reset();
    procedure Change();
    procedure Init(ASize: integer; _A: Tab2DFloat; _b: TabFloat);
    property Results: TResults read FResults;
    property CommandProcessor: TCommandProcessor read FCommandProcessor;
    property Size: integer read FSize write SetSize;
    property Changed: boolean read FChanged;
  end;

  TCommand = class
  private
    FModel: TModel;
  public
    constructor Create(m: TModel);
    function _Do(): boolean; virtual; abstract;
    function Undo(): boolean; virtual; abstract;
    function Describe(): string; virtual; abstract;
    procedure GetColRow(out Col, Row: integer); virtual; abstract;
    procedure GetPosition(out idx1, idx2: integer); virtual; abstract;
    procedure GetValues(out val1, val2: float); virtual; abstract;
  end;

  TCommandProcessor = class
  private
    FModel: TModel;
    Fcommands: TList;
    FcurrentCommand: integer;
    function GetCommandCount: integer;
  public
    constructor Create(m: TModel);
    destructor Destroy(); override;
    function Execute(c: TCommand): boolean;
    function CanUndo(): boolean;
    function CanRedo(): boolean;
    function Undo(): boolean;
    function Redo(): boolean;
    procedure RemoveRedo();
    procedure Clear();
    function Seek(CommandPos: integer): boolean;
    function Describe: string;
    function GetPosition(out idx1, idx2: integer): TClass;
    function GetValues(out val1, val2: float): TClass;
    property CommandCount: integer read GetCommandCount;
    property CurrentCommand: integer read FcurrentCommand;
  end;

  TSetA = class(TCommand)
  private
    Fx, Fy: integer;
    FnewValue, FoldValue: float;
  public
    constructor Create(m: TModel; x, y: integer; Value: float);
    function _Do(): boolean; override;
    function Undo(): boolean; override;
    function Describe(): string; override;
    procedure GetPosition(out idx1, idx2: integer); override;
    procedure GetValues(out val1, val2: float); override;
  end;

  TSetB = class(TCommand)
  private
    Fidx: integer;
    FnewValue, FoldValue: float;
  public
    constructor Create(m: TModel; idx: integer; Value: float);
    function _Do(): boolean; override;
    function Undo(): boolean; override;
    function Describe(): string; override;
    procedure GetPosition(out idx1, idx2: integer); override;
    procedure GetValues(out val1, val2: float); override;
  end;

  TSwapRows = class(TCommand)
  private
    FidxA, FidxB: integer;
  public
    constructor Create(m: TModel; idxA, idxB: integer);
    function _Do(): boolean; override;
    function Undo(): boolean; override;
    function Describe(): string; override;
    procedure GetPosition(out idx1, idx2: integer); override;
    procedure GetValues(out val1, val2: float); override;
  end;

  TSwapColumns = class(TCommand)
  private
    FidxA, FidxB: integer;
  public
    constructor Create(m: TModel; idxA, idxB: integer);
    function _Do(): boolean; override;
    function Undo(): boolean; override;
    function Describe(): string; override;
    procedure GetPosition(out idx1, idx2: integer); override;
    procedure GetValues(out val1, val2: float); override;
  end;

var
  model: TModel;

implementation

uses
  Windows, SysUtils, Math, Forms,
  OptionsUnit;

constructor TCommand.Create(m: TModel);
begin
  FModel := m;
end;

function TCommandProcessor.CanRedo: boolean;
begin
  Result := FcurrentCommand < Fcommands.Count - 1;
end;

function TCommandProcessor.CanUndo: boolean;
begin
  Result := FcurrentCommand >= 0;
end;

procedure TCommandProcessor.Clear;
var
  i: integer;
begin
  for i := Fcommands.Count - 1 downto 0 do
    TCommand(Fcommands[i]).Free();
  Fcommands.Count := 0;
  FcurrentCommand := -1;
end;

constructor TCommandProcessor.Create(m: TModel);
begin
  FModel := m;
  Fcommands := TList.Create();
  FcurrentCommand := -1;
end;

function TCommandProcessor.Describe: string;
begin
  if FcurrentCommand < 0 then
    Result := '(initial state)'
  else
    Result := TCommand(Fcommands[FcurrentCommand]).Describe();
end;

destructor TCommandProcessor.Destroy();
begin
  inherited;
  FcurrentCommand := -1;
  RemoveRedo();
  Fcommands.Free();
end;

function TCommandProcessor.Execute(c: TCommand): boolean;
begin
  Result := c._Do();
  if Result then
  begin
    // remove all "redo" data, because it will be not needed anymore
    if FcurrentCommand < Fcommands.Count - 1 then
      RemoveRedo();
    Assert(Fcommands.Count - 1 = FcurrentCommand);
    Fcommands.Add(c);
    inc(FcurrentCommand);
  end
  else // fail
  begin
    c.Free();
    raise Exception.Create('cammand fail');
  end;
end;

function TCommandProcessor.GetCommandCount: integer;
begin
  Result := Fcommands.Count;
end;

function TCommandProcessor.GetPosition(out idx1, idx2: integer): TClass;
var
  command: TCommand;
begin
  if FcurrentCommand < 0 then
  begin
    idx1 := -1;
    idx2 := -1;
    Result := nil;
  end
  else
  begin
    command := Fcommands[FcurrentCommand];
    Result := command.ClassType;
    command.GetPosition(idx1, idx2);
  end;
end;

function TCommandProcessor.GetValues(out val1, val2: float): TClass;
var
  command: TCommand;
begin
  if FcurrentCommand < 0 then
  begin
    val1 := NAN;
    val2 := NAN;
    Result := nil;
  end
  else
  begin
    command := Fcommands[FcurrentCommand];
    Result := command.ClassType;
    command.GetValues(val1, val2);
  end;
end;

function TCommandProcessor.Undo(): boolean;
begin
  Assert(FcurrentCommand >= 0);
  Result := TCommand(Fcommands[FcurrentCommand]).Undo();
  if Result then
    Dec(FcurrentCommand);
end;

function TCommandProcessor.Redo(): boolean;
begin
  Assert(FcurrentCommand < Fcommands.Count - 1);
  Result := TCommand(Fcommands[FcurrentCommand + 1])._Do();
  if Result then
    inc(FcurrentCommand);
end;

procedure TCommandProcessor.RemoveRedo();
var
  i: integer;
begin
  for i := Fcommands.Count - 1 downto FcurrentCommand + 1 do
    TCommand(Fcommands[i]).Free();
  Fcommands.Count := FcurrentCommand + 1;
end;

function TCommandProcessor.Seek(CommandPos: integer): boolean;
begin
  if (CommandPos < -1) or (CommandPos >= Fcommands.Count) then
  begin
    Result := false;
    exit;
  end;
  if CommandPos = FcurrentCommand then
  begin
    Result := true;
    exit;
  end;
  if CommandPos < FcurrentCommand then
    while FcurrentCommand > CommandPos do
    begin
      Result := Undo();
      if Result = false then
        break;
    end
    else
      while FcurrentCommand < CommandPos do
      begin
        Result := Redo();
        if Result = false then
          break;
      end;
end;

procedure TModel.Change;
begin
  FChanged := true;
end;

constructor TModel.Create();
begin
  inherited;
  FCommandProcessor := TCommandProcessor.Create(Self);
  FResults := TResults.Create(Self);
end;

destructor TModel.Destroy();
begin
  FResults.Free;
  FCommandProcessor.Free;
  inherited;
end;

procedure TModel.Reset;
var
  i: integer;
begin
  SetLength(A, Length(A_sav), Length(A_sav[0]));
  for i := 0 to Length(A_sav) - 1 do
    Move(A_sav[i, 0], A[i, 0], Length(A_sav[i]) * SizeOf(A[i, 0]));
  SetLength(b, Length(b_sav));
  Move(b_sav[0], b[0], Length(b_sav) * SizeOf(b[0]));
end;

procedure TModel.Clear;
var
  i, j: integer;
begin
  for i := 0 to Size - 1 do
    for j := 0 to Size - 1 do
      A[i, j] := 0;
  for i := 0 to Size - 1 do
    b[i] := 0;
  for i := 0 to Size - 1 do
    Results.exact[i] := 0;
  FChanged := true;
end;

procedure TModel.Generate;
var
  i, j: integer;
begin
  { results from -3..3 step 0.25, matrix elements from -5..5 step 1 }
  for i := 0 to Size - 1 do
    for j := 0 to Size - 1 do
      A[i, j] := random(11) - 5;
  for i := 0 to Size - 1 do
    Results.exact[i] := random(25) / 4 - 3;
  for i := 0 to Size - 1 do
  begin
    b[i] := 0;
    for j := 0 to Size - 1 do
      b[i] := b[i] + A[i, j] * Results.exact[j];
  end;
  Results.existsExact := true;
  Results.is_calculated := false;
  FChanged := true;
end;

procedure TModel.Init(ASize: integer; _A: Tab2DFloat; _b: TabFloat);
var
  i: integer;
begin
  SetLength(A, Length(_A), Length(_A[0]));
  for i := 0 to Length(_A) - 1 do
    Move(_A[i, 0], A[i, 0], Length(A[i]) * SizeOf(A[i, 0]));
  SetLength(b, Length(_b));
  Move(_b[0], b[0], Length(b) * SizeOf(b[0]));
  FChanged := true;
end;

procedure TModel.Save;
var
  i: integer;
begin
  SetLength(A_sav, Length(A), Length(A[0]));
  for i := 0 to Length(A) - 1 do
    Move(A[i, 0], A_sav[i, 0], Length(A[i]) * SizeOf(A_sav[i, 0]));
  SetLength(b_sav, Length(b));
  Move(b[0], b_sav[0], Length(b) * SizeOf(b_sav[0]));
  FChanged := false;
end;

procedure TModel.SetSize(const Value: integer);
begin
  if Value > 1000 then
    FSize := 1000
  else
    FSize := Value;
  SetLength(A, FSize, FSize);
  SetLength(b, FSize);
  FResults.Size := Size;
end;

constructor TSetA.Create(m: TModel; x, y: integer; Value: float);
begin
  inherited Create(m);
  Fx := x;
  Fy := y;
  FoldValue := NAN; // not disturb that FoldValue has no old value, it will set in _Do()
  FnewValue := Value;
end;

function TSetA._Do(): boolean;
begin
  FoldValue := FModel.A[Fx, Fy];
  FModel.A[Fx, Fy] := FnewValue;
  Result := true;
end;

function TSetA.Undo(): boolean;
begin
  FnewValue := FModel.A[Fx, Fy];
  FModel.A[Fx, Fy] := FoldValue;
  Result := true;
end;

function TSetA.Describe(): string;
begin
  Result := Format
    ('Set A[%d,%d] to ' + Opt.FormatStr + ' (was ' + Opt.FormatStr + ')',
    [Fx, Fy, FnewValue, FoldValue]);
end;

procedure TSetA.GetPosition(out idx1, idx2: integer);
begin
  idx1 := Fx;
  idx2 := Fy;
end;

procedure TSetA.GetValues(out val1, val2: float);
begin
  val1 := FoldValue;
  val2 := FnewValue;
end;

constructor TSetB.Create(m: TModel; idx: integer; Value: float);
begin
  inherited Create(m);
  Fidx := idx;
  FoldValue := NAN;
  FnewValue := Value;
end;

function TSetB._Do(): boolean;
begin
  FoldValue := FModel.b[Fidx];
  FModel.b[Fidx] := FnewValue;
  Result := true;
end;

function TSetB.Undo(): boolean;
begin
  FnewValue := FModel.b[Fidx];
  FModel.b[Fidx] := FoldValue;
  Result := true;
end;

function TSetB.Describe(): string;
begin
  Result := Format('Set B[%d] to %.4f (was %.4f)', [Fidx, FnewValue, FoldValue]
    );
end;

procedure TSetB.GetPosition(out idx1, idx2: integer);
begin
  idx1 := Fidx;
  idx2 := -1;
end;

procedure TSetB.GetValues(out val1, val2: float);
begin
  val1 := FoldValue;
  val2 := FnewValue;
end;

{ TSwapRows }

constructor TSwapRows.Create(m: TModel; idxA, idxB: integer);
begin
  inherited Create(m);
  FidxA := idxA;
  FidxB := idxB;
end;

function TSwapRows.Describe: string;
begin
  Result := Format('Swap rows %d and %d', [FidxA, FidxB]);
end;

procedure TSwapRows.GetPosition(out idx1, idx2: integer);
begin
  idx1 := FidxA;
  idx2 := FidxB;
end;

procedure TSwapRows.GetValues(out val1, val2: float);
begin
  val1 := 0;
  val2 := 0;
end;

function TSwapRows.Undo: boolean;
begin
  Result := _Do();
end;

function TSwapRows._Do: boolean;
var
  tmp: float;
  i: integer;
begin
  with model do
  begin
    for i := 0 to Size - 1 do
    begin
      tmp := A[FidxA, i];
      A[FidxA, i] := A[FidxB, i];
      A[FidxB, i] := tmp;
    end;
    tmp := b[FidxA];
    b[FidxA] := b[FidxB];
    b[FidxB] := tmp;
  end;
  Result := true;
end;

{ TSwapColumns }

constructor TSwapColumns.Create(m: TModel; idxA, idxB: integer);
begin
  inherited Create(m);
  FidxA := idxA;
  FidxB := idxB;
end;

function TSwapColumns.Describe: string;
begin
  Result := Format('Swap columns %d and %d', [FidxA, FidxB]);
end;

procedure TSwapColumns.GetPosition(out idx1, idx2: integer);
begin
  idx1 := FidxA;
  idx2 := FidxB;
end;

procedure TSwapColumns.GetValues(out val1, val2: float);
begin
  val1 := 0;
  val2 := 0;
end;

function TSwapColumns.Undo: boolean;
begin
  Result := _Do();
end;

function TSwapColumns._Do: boolean;
var
  tmp: float;
  i: integer;
begin
  with model do
    for i := 0 to Size - 1 do
    begin
      tmp := A[i, FidxA];
      A[i, FidxA] := A[i, FidxB];
      A[i, FidxB] := tmp;
    end;
  Result := true;
end;

{ TResults }

procedure TResults.CopyExact(Aexact: TabFloat);
begin
  SetLength(exact, Length(Aexact));
  if Length(exact) > 0 then
    Move(Aexact[0], exact[0], Length(exact) * SizeOf(exact[0]));
end;

constructor TResults.Create(AModel: TModel);
begin
  inherited Create;
  FModel := AModel;
end;

procedure TResults.RestoreOrder;
var
  i: integer;
  tmp: float;
begin
  for i := FModel.Size - 2 downto 0 do
    if swaps[i] <> 0 then
    begin
      tmp := calculated[i];
      calculated[i] := calculated[swaps[i]];
      calculated[swaps[i]] := tmp;
    end;
end;

procedure TResults.SetSize(const Value: integer);
begin
  FSize := Value;
  SetLength(exact, FSize);
  SetLength(calculated, FSize);
  SetLength(swaps, FSize - 1);
end;

initialization

model := TModel.Create();

finalization

model.Free;

end.
